from .robot import Robot


class Eddy(object):
    def __init__(self, host=None):
        if host is None:
            hosts = Robot.discover()['eddy']
            if len(hosts) != 1:
                raise IOError('Eddy not found! {}'.format(hosts))
            host = hosts[0]

        self._robot = Robot(host)

        self.left_wheel = self._robot.eddy.left_wheel
        self.right_wheel = self._robot.eddy.right_wheel

        self.left_light_sensor = self._robot.eddy.right_light_sensor
        self.right_light_sensor = self._robot.eddy.left_light_sensor



import math
import time

from .robot import Robot


class Finger(object):
    def __init__(self, m):
        self._m = m
        self._flex = None

    @property
    def flex(self):
        return self._flex

    @flex.setter
    def flex(self, val):
        self._flex = val
        val = min(max(val, 0), 100)
        val = val * 1.65
        self._m.target_position = int(val)


class Handy(object):
    def __init__(self, host=None):
        if host is None:
            hosts = Robot.discover()['handy']
            if len(hosts) != 1:
                raise IOError('Handy not found! {}'.format(hosts))
            host = hosts[0]

        self._robot = Robot(host)

        self.index = Finger(self._robot.handy.m1)
        self.middle_finger = Finger(self._robot.handy.m2)
        self.ring_finger = Finger(self._robot.handy.m3)
        self.pinky = Finger(self._robot.handy.m4)
        self.thumb = Finger(self._robot.handy.m5)

        self.fingers = [
            self.thumb,
            self.index,
            self.middle_finger,
            self.ring_finger,
            self.pinky,
        ]

    def sync_wave(self):
        t0 = time.time()

        while True:
            t = time.time() - t0
            target = 50 + 50 * math.sin(1 * t)

            for f in self.fingers:
                f.flex = target

    def wave(self):
        while True:
            t = time.time()
            for i, f in enumerate(self.fingers):
                f.flex = 50 + 50 * math.sin(1 * t + i * math.pi / 5)



from tornado.ioloop import IOLoop
from tornado.web import Application, RequestHandler
from tornado.websocket import WebSocketHandler

from pyluos import Robot
from pyluos.io import IOHandler
from pyluos.tools import scratch


class UsbHandler(IOHandler):
    def __init__(self, host):
        self.robot = Robot(host)
        self.io = self.robot._io

        self.ws = None

    def write(self, msg):
        self.io.write(msg)

    def is_ready(self):
        return self.io.is_ready()

    def recv(self):
        message = self.io.recv()
        if self.ws is not None:
            self.ws.ioloop.add_callback(lambda: self.ws.write_message(message))
        return message


class WsHandler(WebSocketHandler):
    def open(self):
        if self.robot._io.ws is not None:
            self.robot._io.ws.close()

        self.robot._io.ws = self

    def on_close(self):
        self.robot._io.ws = None

    def on_message(self, message):
        self.robot._io.write(message.encode())

    def check_origin(self, origin):
        return True


class HttpScratchHandler(RequestHandler):
    def get(self):
        self.write(self.extension)


class Test(RequestHandler):
    def get(self):
        self.write("hello")


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('serial_port', type=str)
    parser.add_argument('--port', type=int, default=9342)
    parser.add_argument('--hostname', type=str, default='localhost')
    parser.add_argument('--verbose', action='store_true', default=False)
    args = parser.parse_args()

    robot = Robot(args.serial_port, IO=UsbHandler)

    loop = IOLoop()
    app = Application([
        (r'/', WsHandler),
        (r'/scratch-extension', HttpScratchHandler),
    ])

    WsHandler.robot = robot
    WsHandler.ioloop = loop
    WsHandler.verbose = args.verbose

    ext = scratch.generate_extension('luos', robot, args.hostname, args.port)
    HttpScratchHandler.extension = ext

    print('Scratch extension url: "http://{}:{}/scratch-extension"'.format(args.hostname, args.port))

    app.listen(args.port)
    loop.start()


if __name__ == '__main__':
    main()



from .module import Module


class Void(Module):

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Void', id, alias, robot)

    def _update(self, new_state):
        Module._update(self, new_state)

    def dxl_detect(self):
        self._push_value('reinit', 0)



from .module import Module, interact

class DynamixelMotor(Module):
    def __init__(self, id, alias, robot):
        Module.__init__(self, 'DynamixelMotor', id, alias, robot)
        # Read
        self.position = None
        self.temperature = None

        # Write
        self._target_position = None
        self._moving_speed = None
        self._compliant = None
        self._wheel_mode = None

    def _update(self, new_state):
        Module._update(self, new_state)

        if 'rot_position' in new_state:
            self.position = new_state['rot_position']
        if 'temperature' in new_state:
            self.temperature = new_state['temperature']

    @property
    def target_position(self):
        return self._target_position

    @target_position.setter
    def target_position(self, target_position):
        if target_position != self._target_position and self._compliant == False:
            self._push_value('target_rot_position', target_position)
            self._target_position = target_position

    @property
    def moving_speed(self):
        return self._moving_speed

    @moving_speed.setter
    def moving_speed(self, moving_speed):
        if moving_speed != self._moving_speed:
            self._push_value('target_rot_speed', moving_speed)
            self._moving_speed = moving_speed

    @property
    def compliant(self):
        return self._compliant

    @compliant.setter
    def compliant(self, compliant):
        self._push_value('compliant', compliant)
        self._compliant = compliant

    @property
    def wheel_mode(self):
        return self._wheel_mode

    @wheel_mode.setter
    def wheel_mode(self, wheel_mode):
        if wheel_mode != self._wheel_mode:
            self._push_value('wheel_mode', wheel_mode)
            self._wheel_mode = wheel_mode

    def detect(self):
        self._push_value('reinit', 0)

    def register(self, register, val):
        new_val = [register, val]
        self._push_value('register', new_val)

    # notebook things
    def control(self):
        def change_position(target_position):
            self.target_position = target_position

        return interact(change_position, target_position=(-150.0, 150.0, 1.0))



import os
import sys
import json
import time
import logging
import threading
import logging.config

from datetime import datetime
from collections import defaultdict

from .io import discover_hosts, io_from_host, Ws
from .modules import name2mod


def run_from_unittest():
    return 'unittest' in sys.modules


use_topographe = True
try:
    import zmq
except ImportError:
    use_topographe = False


known_host = {
    'ergo': ['/dev/cu.usbserial-DN2AAOVK', '/dev/cu.usbserial-DN2YEFLN'],
    'handy': ['/dev/cu.usbserial-DN2X236E'],
    'eddy': ['pi-gate.local'],
}


class Robot(object):
    _heartbeat_timeout = 5  # in sec.
    _max_alias_length = 15
    _base_log_conf = os.path.join(os.path.dirname(__file__),
                                  'logging_conf.json')

    @classmethod
    def discover(cls):
        hosts = discover_hosts()

        possibilities = {
            k: [h for h in v if h in hosts]
            for k, v in known_host.items()
        }

        return possibilities

    def __init__(self, host,
                 IO=None,
                 log_conf=_base_log_conf,
                 test_mode=False,
                 *args, **kwargs):
        if IO is not None:
            self._io = IO(host=host, *args, **kwargs)
        else:
            self._io = io_from_host(host=host,
                                    *args, **kwargs)

        if os.path.exists(log_conf):
            with open(log_conf) as f:
                config = json.load(f)
            logging.config.dictConfig(config)

        self.logger = logging.getLogger(__name__)
        self.logger.info('Connected to "{}".'.format(host))

        self._send_lock = threading.Lock()
        self._cmd_lock = threading.Lock()

        # We force a first poll to setup our model.
        self._setup()
        self.logger.info('Robot setup.')

        self._last_update = time.time()
        self._running = True

        # Setup both poll/push synchronization loops.
        self._poll_bg = threading.Thread(target=self._poll_and_up)
        self._poll_bg.daemon = True
        self._poll_bg.start()

        if not (test_mode or run_from_unittest()) and use_topographe:
            c = zmq.Context()
            s = c.socket(zmq.PUB)
            s.connect('tcp://127.0.0.1:33000')
            self._s = s

    @property
    def state(self):
        return {
            'gate': self.name,
            'timestamp': datetime.now(),
            'types': ','.join([mod.type for mod in self.modules]),
            'modules': ','.join([mod.alias for mod in self.modules])
        }

    @property
    def name(self):
        return self._name

    @property
    def alive(self):
        dt = time.time() - self._last_update
        return self._running and dt < self._heartbeat_timeout

    def close(self):
        self._running = False
        self._poll_bg.join()
        self._io.close()

    def _setup(self):
        self.logger.info('Sending detection signal.')
        self._send({'detection': {}})
        if isinstance(self._io, Ws):
            import time
            time.sleep(0.1)
            self._send({'detection': {}})

        self.logger.info('Waiting for route table...')

        state = self._poll_once()
        while ('route_table' not in state):
            state = self._poll_once()

        print (state)

        try:
            gate = next(g for g in state['route_table']
                        if 'type' in g and g['type'] == 'gate')
            self._name = gate['alias']
        except StopIteration:
            self._name = 'gate_unknown'

        modules = [mod for mod in state['route_table']
                   if 'type' in mod and mod['type'] in name2mod.keys()]

        self._cmd = defaultdict(lambda: defaultdict(lambda: None))

        self.modules = [
            name2mod[mod['type']](id=mod['id'],
                                  alias=mod['alias'],
                                  robot=self)
            for mod in modules
            if 'type' in mod and 'id' in mod and 'alias' in mod
        ]

        for mod in self.modules:
            setattr(self, mod.alias, mod)

        # We push our current state to make sure that
        # both our model and the hardware are synced.
        self._push_once()

    # Poll state from hardware.
    def _poll_once(self):
        self._state = self._io.read()
        self._state['timestamp'] = time.time()
        return self._state

    def _poll_and_up(self):
        while self._running:
            state = self._poll_once()
            self._update(state)
            self._push_once()

            self._broadcast(state)

    # Update our model with the new state.
    def _update(self, new_state):
        if 'modules' not in new_state:
            return

        for alias, mod in new_state['modules'].items():
            if hasattr(self, alias):
                getattr(self, alias)._update(mod)

        self._last_update = time.time()

    def update_cmd(self, alias, key, val):
        with self._cmd_lock:
            self._cmd[alias][key] = val

    def _push_once(self):
        with self._cmd_lock:
            if not self._cmd:
                return

            self._send({
                'modules': self._cmd
            })

            self._cmd = defaultdict(lambda: defaultdict(lambda: None))

    def _send(self, msg):
        with self._send_lock:
            self._io.send(msg)

    def _broadcast(self, state):
        if not hasattr(self, '_s'):
            return

        for mod in state['modules']:
            if ((mod['type'] == 'servo') and
               (hasattr(self, mod['alias']))):
                servo = getattr(self, mod['alias'])
                mod['position'] = servo.target_position

        msg = '{} {}'.format(self.name, json.dumps(state))
        self._s.send_string(msg)

    def rename_module(self, old, new):
        if not hasattr(self, old):
            raise ValueError('No module named {}!'.format(old))

        if len(new) > self._max_alias_length:
            raise ValueError(
                'Alias length should be less than {}'
                .format(self._max_alias_length)
            )

        self._send({'modules': {old: {'set_alias': new}}})

        mod = getattr(self, old)
        mod.alias = new
        setattr(self, new, mod)
        delattr(self, old)



from .robot import Robot


class ErgoJr(object):
    def __init__(self, host=None):
        if host is None:
            hosts = Robot.discover()['ergo']
            if len(hosts) != 1:
                raise IOError('ErgoJr not found! {}'.format(hosts))
            host = hosts[0]

        self._robot = Robot(host)

        self._dxl = self._robot.dxl
        while not self._dxl._setup:
            pass
        setattr(self, 'motors', self._dxl.motors)
        for i, m in enumerate(self._dxl.motors):
            setattr(self, 'm{}'.format(i + 1), m)

        for mod in self._robot.modules:
            if mod.type != 'Dynamixel':
                setattr(self, mod.alias, mod)



import os
import re

from jinja2 import Template

DEFAULT_TEMPLATE = os.path.join(os.path.dirname(__file__), 'extension.tpl.js')

supported_modules = [
    'button',
    'dynamixel',
    'l0_dc_motor',
    'l0_gpio',
    'l0_servo',
    'led',
    'potard',
]


def find_modules(state, type):
    return [
        str(m['alias'])
        for m in state['modules']
        if m['type'] == type
    ]


def find_xl320(state, dxl):
    dxl = next(filter(lambda mod: mod['alias'] == dxl, state['modules']))
    motors = [m for m in dxl if re.match(r'm[0-9]+', m)]
    return motors


def generate_extension(name,
                       robot,
                       host,
                       port,
                       template=DEFAULT_TEMPLATE):
    context = {}
    context = {
        'name': name,
        'host': host,
        'port': port,
    }
    context.update({
        type: find_modules(robot._state, type)
        for type in supported_modules
    })

    if context['dynamixel']:
        # TODO: This should be done for every dxl controller!
        context['xl_320'] = find_xl320(robot._state, context['dynamixel'][0])

    with open(template) as f:
        tpl = Template(f.read())
        ext = tpl.render(**context)
        return ext



from .module import Module


class Potentiometer(Module):
    possible_events = {'moved'}

    def __init__(self, id, alias, robot):
        Module.__init__(self, 'Potentiometer', id, alias, robot)
        self._value = 0

    @property
    def position(self):
        """ Position in degrees. """
        return self._value

    def _update(self, new_state):
        Module._update(self, new_state)
        if 'rot_position' in new_state:
            new_pos = new_state['rot_position']
            if new_pos != self._value:
                self._value = new_pos
                self._pub_event('moved', self._value, self.position)



import logging

from .robot import Robot
from .modules import *

from .eddy import Eddy
from .ergo import ErgoJr
from .handy import Handy

nh = logging.NullHandler()
logging.getLogger(__name__).addHandler(nh)
